import { INDENT, INTERNAL, formatType, toHex } from './common';

export class Method {
    readonly self: string;
    readonly shortName: string;
    readonly returnType: string;
    readonly args: string;
    readonly address: number;
    readonly isInherited: boolean;
    readonly isStatic: boolean;

    // Constructor
    constructor(self: string, name: string, returnType: string, args: string, address: number, isInherited: boolean, isStatic: boolean) {
        this.self = self;
        this.shortName = name;
        this.returnType = returnType;
        this.args = args;
        this.address = address;
        this.isInherited = isInherited;
        this.isStatic = isStatic;
    }

    // Getters
    getName(separator: string = '_') {
        return this.self + separator + this.shortName;
    }
    getType() {
        return this.getName() + '_t';
    }
    getProperty(hasWrapper: boolean) {
        let out = INDENT;
        if (hasWrapper) {
            out += `${this.getWrapperType()}::ptr_type ${this.shortName}`;
        } else {
            out += `${formatType(this.returnType.trim())}(*${this.shortName})${this.args.trim()}`;
        }
        out += ';\n';
        return out;
    }
    getWrapperType() {
        return `std::remove_pointer_t<decltype(${this.getName()})>`;
    }
    #getSignature() {
        return this.returnType.trim() + this.args.trim();
    }
    #getFullType() {
        return `${INTERNAL}Function<${this.#getSignature()}>`;
    }

    // Overwrite Helper
    #getVirtualCall(self: string = this.self) {
        return `${self}_vtable_base->${this.shortName}`;
    }
    generate(code: boolean, isVirtual: boolean, parentSelf?: string) {
        let out = '';
        if (!code) {
            out += 'extern ';
        }
        const type = this.#getFullType();
        out += `${type} *const ${this.getName()}`;
        if (code) {
            out += ` = new ${type}(${JSON.stringify(this.getName('::'))}, `;
            if (isVirtual) {
                const parentMethod = parentSelf ? `(void *) ${this.#getVirtualCall(parentSelf)}` : 'nullptr';
                out += `&${this.#getVirtualCall()}, ${parentMethod}`;
            } else {
                out += `${this.getWrapperType()}::ptr_type(${toHex(this.address)})`;
            }
            out += `, ${INTERNAL}thunk<&${this.getName()}>)`;
        }
        out += ';\n';
        if (!code) {
            out += `typedef ${this.getWrapperType()}::type ${this.getType()};\n`;
        }
        return out;
    }
}